from typing import cast

from pybotx_smartapp_rpc.empty_args import EmptyArgs
from pybotx_smartapp_rpc.smartapp import SmartApp
from pybotx_smartapp_rpc.typing import (
    Handler,
    HandlerWithArgs,
    HandlerWithoutArgs,
    RPCArgsBaseModel,
    RPCResponse,
)


async def empty_args_middleware(
    smartapp: SmartApp,
    rpc_arguments: RPCArgsBaseModel,
    call_next: Handler,
) -> RPCResponse:
    if isinstance(rpc_arguments, EmptyArgs):
        call_next = cast(HandlerWithoutArgs, call_next)
        return await call_next(smartapp)

    call_next = cast(HandlerWithArgs, call_next)
    return await call_next(smartapp, rpc_arguments)



from typing import Awaitable, Callable, Dict, Type, TypeVar, Union

from pydantic import BaseModel as RPCArgsBaseModel

from pybotx_smartapp_rpc.models.responses import RPCErrorResponse, RPCResultResponse
from pybotx_smartapp_rpc.smartapp import SmartApp

RPCResponse = Union[RPCErrorResponse, RPCResultResponse]

TArgs = TypeVar("TArgs", bound=RPCArgsBaseModel)

HandlerWithArgs = Callable[[SmartApp, TArgs], Awaitable[RPCResponse]]
HandlerWithoutArgs = Callable[[SmartApp], Awaitable[RPCResponse]]
Handler = Union[HandlerWithArgs, HandlerWithoutArgs]
Middleware = Callable[[SmartApp, TArgs, HandlerWithArgs], Awaitable[RPCResponse]]

TException = TypeVar("TException", bound=Exception)
ExceptionHandler = Callable[[TException, SmartApp], Awaitable[RPCErrorResponse]]
ExceptionHandlerDict = Dict[Type[Exception], ExceptionHandler]



from typing import List, Optional

from pybotx import Bot, SmartAppEvent
from pydantic.error_wrappers import ValidationError

from pybotx_smartapp_rpc.exception_handlers import (
    default_exception_handler,
    rpc_exception_handler,
)
from pybotx_smartapp_rpc.exceptions import RPCErrorExc
from pybotx_smartapp_rpc.middlewares.exception_middleware import ExceptionMiddleware
from pybotx_smartapp_rpc.models.request import RPCRequest
from pybotx_smartapp_rpc.models.responses import (
    build_invalid_rpc_request_error_response,
)
from pybotx_smartapp_rpc.router import RPCRouter
from pybotx_smartapp_rpc.smartapp import SmartApp
from pybotx_smartapp_rpc.typing import ExceptionHandlerDict, Middleware, RPCResponse


class SmartAppRPC:
    def __init__(
        self,
        routers: List[RPCRouter],
        middlewares: Optional[List[Middleware]] = None,
        exception_handlers: Optional[ExceptionHandlerDict] = None,
    ) -> None:
        self._middlewares = middlewares or []
        self._insert_exception_middleware(exception_handlers or {})

        self._router = self._merge_routers(routers)

    async def handle_smartapp_event(self, event: SmartAppEvent, bot: Bot) -> None:
        rpc_response: RPCResponse

        try:
            rpc_request = RPCRequest(**event.data)
        except ValidationError as invalid_rcp_request_exc:
            rpc_response = build_invalid_rpc_request_error_response(
                invalid_rcp_request_exc,
            )
        else:
            rpc_response = await self._router.perform_rpc_request(
                SmartApp(bot, event.bot.id, event.chat.id, event),
                rpc_request,
            )

        await bot.send_smartapp_event(
            bot_id=event.bot.id,
            chat_id=event.chat.id,
            data=rpc_response.jsonable_dict(),
            ref=event.ref,
            files=rpc_response.files,
        )

    def _insert_exception_middleware(
        self,
        user_exception_handlers: ExceptionHandlerDict,
    ) -> None:
        exception_handlers: ExceptionHandlerDict = {
            Exception: default_exception_handler,
            RPCErrorExc: rpc_exception_handler,
        }
        exception_handlers.update(user_exception_handlers)
        exc_middleware = ExceptionMiddleware(exception_handlers)
        self._middlewares.insert(0, exc_middleware)

    def _merge_routers(self, routers: List[RPCRouter]) -> RPCRouter:
        main_router = RPCRouter(middlewares=self._middlewares)
        main_router.include(*routers)

        return main_router



from types import SimpleNamespace
from typing import Any, List, Optional
from uuid import UUID

from pybotx import Bot, File, SmartAppEvent
from pybotx.missing import Missing, Undefined


class SmartApp:
    def __init__(
        self,
        bot: Bot,
        bot_id: UUID,
        chat_id: UUID,
        event: Optional[SmartAppEvent] = None,
    ) -> None:
        self.bot = bot
        self.event = event

        self.bot_id = bot_id
        self.chat_id = chat_id

        self.state = SimpleNamespace()

    async def send_event(
        self,
        rpc_result: Any,
        files: Optional[List[File]] = None,
    ) -> None:
        await self.bot.send_smartapp_event(
            bot_id=self.bot_id,
            chat_id=self.chat_id,
            data={
                "status": "ok",
                "type": "smartapp_rpc",
                "result": rpc_result,
            },
            files=files or [],
        )

    async def send_push(self, counter: int, body: Missing[str] = Undefined) -> None:
        await self.bot.send_smartapp_notification(
            bot_id=self.bot_id,
            chat_id=self.chat_id,
            smartapp_counter=counter,
            body=body,
        )



from loguru import logger

from pybotx_smartapp_rpc.exceptions import RPCErrorExc
from pybotx_smartapp_rpc.models.errors import RPCError
from pybotx_smartapp_rpc.models.responses import RPCErrorResponse
from pybotx_smartapp_rpc.smartapp import SmartApp


async def default_exception_handler(
    exc: Exception,
    smartapp: SmartApp,
) -> RPCErrorResponse:
    logger.exception(exc)
    return RPCErrorResponse(
        errors=[RPCError(reason="Internal error", id=exc.__class__.__name__.upper())],
    )


async def rpc_exception_handler(
    exc: RPCErrorExc,
    smartapp: SmartApp,
) -> RPCErrorResponse:
    return RPCErrorResponse(errors=exc.errors)



from dataclasses import asdict, dataclass, field
from typing import Any, Dict, Generic, List, TypeVar

from pybotx import File
from pydantic.error_wrappers import ValidationError

from pybotx_smartapp_rpc.models.errors import RPCError

ResultType = TypeVar("ResultType")


@dataclass
class RPCResultResponse(Generic[ResultType]):
    result: ResultType
    files: List[File] = field(default_factory=list)

    def jsonable_dict(self) -> Dict[str, Any]:
        return {
            "status": "ok",
            "type": "smartapp_rpc",
            "result": self.result,
        }


@dataclass
class RPCErrorResponse:
    errors: List[RPCError]
    files: List[File] = field(default_factory=list)

    def jsonable_dict(self) -> Dict[str, Any]:
        return {
            "status": "error",
            "type": "smartapp_rpc",
            "errors": [asdict(error) for error in self.errors],
        }


def build_invalid_rpc_request_error_response(
    exc: ValidationError,
) -> RPCErrorResponse:
    return RPCErrorResponse(
        errors=[
            RPCError(
                reason=f"Invalid RPC request: {error['msg']}",
                id=error["type"].split(".")[0].upper(),
                meta={"field": error["loc"][0]},
            )
            for error in exc.errors()
        ],
    )


def build_invalid_rpc_args_error_response(
    exc: ValidationError,
) -> RPCErrorResponse:
    return RPCErrorResponse(
        errors=[
            RPCError(
                reason=error["msg"],
                id=error["type"].split(".")[0].upper(),
                meta={"field": error["loc"][0]},
            )
            for error in exc.errors()
        ],
    )


def build_method_not_found_error_response(
    method: str,
) -> RPCErrorResponse:
    return RPCErrorResponse(
        errors=[
            RPCError(
                reason="Method not found",
                id="METHOD_NOT_FOUND",
                meta={"method": method},
            ),
        ],
    )



from typing import List, Union

from pybotx_smartapp_rpc.models.errors import RPCError


class BaseRPCErrorExc(Exception):
    ...  # noqa: WPS428, WPS604


class RPCErrorExc(BaseRPCErrorExc):
    def __init__(self, error_or_errors: Union[RPCError, List[RPCError]]):
        if isinstance(error_or_errors, RPCError):
            self.errors = [error_or_errors]
        else:
            self.errors = error_or_errors



from dataclasses import dataclass
from functools import partial
from typing import List, Optional, Type

from pybotx_smartapp_rpc.smartapp import SmartApp
from pybotx_smartapp_rpc.typing import (
    Handler,
    HandlerWithArgs,
    Middleware,
    RPCArgsBaseModel,
    RPCResponse,
)


@dataclass
class RPCMethod:
    handler: Handler
    middlewares: List[Middleware]
    arguments_class: Optional[Type[RPCArgsBaseModel]] = None

    async def __call__(
        self,
        smartapp: SmartApp,
        rpc_args: RPCArgsBaseModel,
    ) -> RPCResponse:
        # loop in reverse order
        # if middlewares = [m1, m2] and method.middlewares = [m3, m4]
        # then stack will be m1(m2(m3(m4(handler()))))
        handler: HandlerWithArgs = self.handler  # type: ignore
        for middleware in self.middlewares[::-1]:
            part = partial(middleware, call_next=handler)
            handler = part

        return await handler(smartapp, rpc_args)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pybotx_smartapp_rpc',
 'pybotx_smartapp_rpc.middlewares',
 'pybotx_smartapp_rpc.models']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0.70.0,<0.75.0', 'pybotx>=0.32.0,<0.45.0', 'pydantic>=1.8.2,<2.0.0']

setup_kwargs = {
    'name': 'pybotx-smartapp-rpc',
    'version': '0.4.4',
    'description': 'eXpress SmartApp JSON-RPC library',
    'long_description': '# BotX-SmartApp-RPC\nБиблиотека, позволяющая писать смартаппы, используя [наш JSONRPC-like протокол](https://ccsteam.atlassian.net/wiki/spaces/EI/pages/193167368/SmartApp+RPC)\n\n## Установка\nИспользуя `poetry`:\n\n```bash\npoetry add pybotx-smartapp-rpc\n```\n\n## Добавление RPC методов\n1. Создайте класс для входящих аргументов:\n``` python\nfrom pybotx_smartap_rpc import RPCArgsBaseModel\n...\nclass SumArgs(RPCArgsBaseModel):\n    a: int\n    b: int\n```\n2. Создайте RPC метод:\n``` python\nfrom pybotx_smartapp_rpc import SmartApp, RPCRouter, RPCResultResponse\n...\nrpc = RPCRouter()\n...\n@rpc.method("sum")\nasync def sum(\n    smartapp: SmartApp, rpc_arguments: SumArgs\n) -> RPCResultResponse[int]:\n    return RPCResultResponse(result=rpc_arguments.a + rpc_arguments.b)\n\n# Так же у метода может не быть аргументов:\n@rpc.method("answer")\nasync def answer(smartapp: SmartApp) -> RPCResultResponse[int]:\n    return RPCResultResponse(result=42)\n```\n3. Создайте экземпляр `SmartAppRPC` и подключите роутер из прошлого пункта:\n``` python\nfrom pybotx_smartapp_rpc import SmartAppRPC\n\nfrom anywhere import methods \n...\nsmartapp = SmartAppRPC(routers=[methods.rpc])\n```\n4. Сделайте хендлер для `smartapp_event` и вызывайте в нем хендлер библиотеки\n``` python\n@collector.smartapp_event\nasync def handle_smartapp_event(event: SmartAppEvent, bot: Bot) -> None:\n    await smartapp.handle_smartapp_event(event, bot)\n```\n\n## Продвинутая работа с библиотекой\n* В `RPCResultResponse` можно передавать `botx.File` файлы.\n``` python\n@rpc.method("get-pdf")\nasync def get_pdf(\n    smartapp: SmartApp, rpc_arguments: GetPDFArgs\n) -> RPCResultResponse[None]:\n    ...\n    return RPCResultResponse(result=None, files=[...])\n```\n* В `SmartAppRPC`, `RPCRouter` и `RPCRouter.method` можно передать мидлвари, сначала будут вызваны мидлвари приложения, затем мидлвари роутера и в конце мидлвари метода.\n``` python\nsmartapp = SmartAppRPC(..., middlewares=[...])\n...\nrpc = RPCRouter(middlewares=[...])\n...\n@rpc.method("sum", middlewares=[...])\n```\n* `RPCArgsBaseModel` это алиас для `pydantic.BaseModel`, вы можете использовать все возможности исходного класса.\n``` python\nfrom uuid import UUID\n...\nclass DelUserArgs(RPCArgsBaseModel):\n    # pydantic сериализует входящую строку в UUID\n    user_huid: UUID\n```\n* Через объект `smartapp`, передаваемый в хендлер можно получить доступ к `event` и `bot`.\n``` python\n...\n@rpc.method("del-user")\nasync def del_user(\n    smartapp: SmartApp, rpc_arguments: DelUserArgs\n) -> RPCResultResponse[None]:\n    await smartapp.bot.send_message(\n        body="Done",\n        bot_id=smartapp.event.bot.id,\n        chat_id=smartapp.event.chat.id,\n    )\n    ...\n```\n* Используя метод `smartapp.send_event` можно отправлять RPC ивенты с `ref: null`.  \nЭто может пригодиться при необходимости отправки уведомления не в ответ на RPC запрос.\n``` python\n@rpc.method("notify-me")\nasync def notify_me(\n    smartapp: SmartApp, rpc_arguments: NotifyMeArgs\n) -> RPCResultResponse[None]:\n    ...\n    await smartapp.send_event("notified", files=[notify_file])\n    ...\n```\n* Используя метод `smartapp.send_push` можно отправлять пуш уведомлений на клиент.\nИ обновлять счетчик уведомлений на икноке смартапа.\n``` python\n@rpc.method("notify-me")\nasync def notify_me(\n    smartapp: SmartApp, rpc_arguments: NotifyMeArgs\n) -> RPCResultResponse[None]:\n    await smartapp.send_push(42, "You have 42 new emails!")\n    ...\n```\n* В мидлварях можно создавать новые объекты в `smartapp.state`, чтобы потом использовать их в хендлерах.\n``` python\nasync def user_middleware(smartapp: SmartApp, rpc_arguments: RPCArgsBaseModel, call_next: Callable) -> RPCResponse[User]:\n    smartapp.state.user = await User.get(smartapp.message.user_huid)\n    return await call_next(smartapp, rpc_arguments)\n\n@rpc.method("get-user-fullname")\nasync def get_user_fullname(smartapp: SmartApp) -> RPCResultResponse[str]:\n    return RPCResultResponse(result=smartapp.state.user.fullname)\n```\n* Можно выбрасывать пользовательские RPC ошибки, которые будут отправлены как ответ на RPC запрос.\n``` python\nfrom pybotx_smartapp_rpc import RPCErrorExc, RPCError\n...\n@rpc.method("return-error")\nasync def return_error(smartapp: SmartApp, rpc_arguments: RaiseOneErrorArgs) -> None:\n    # one error\n    raise RPCErrorExc(\n        RPCError(\n            reason="It\'s error reason",\n            id="CUSTOM_ERROR",\n            meta={"args": rpc_arguments.dict()},\n        )\n    )\n    # or list of errors\n    raise RPCErrorExc(\n        [\n            RPCError(\n                reason="It\'s error reason",\n                id="CUSTOM_ERROR",\n                meta={"args": rpc_arguments.dict()},\n            ),\n            RPCError(\n                reason="It\'s one more error reason",\n                id="CUSTOM_ERROR_NUMBER_TWO",\n                meta={"args": rpc_arguments.dict()},\n            )\n        ]\n    )\n```\n* Можно добавить хендлер на определенный тип исключений. В него будут отправлять исключения того же и дочерних классов.\nХендлер **обязан** возвращать `RPCErrorResponse`, ошибки из которого будут отправлены источнику запроса.\n``` python\nfrom pybotx_smartapp_rpc import SmartAppRPC, RPCErrorResponse\n...\nasync def key_error_handler(exc: KeyError, smartapp: SmartApp) -> RPCErrorResponse:\n    key = exc.args[0]\n    return RPCErrorResponse(\n        errors=[\n            RPCError(\n                reason=f"Key {key} not found.",\n                id="KEY_ERROR",\n                meta={"key": key},\n            ),\n        ]\n    )\n\nsmartapp = SmartAppRPC(..., exception_handlers={KeyError: key_error_handler})\n```\n',
    'author': 'Arseniy Zhiltsov',
    'author_email': 'arseniy.zhiltsov@ccsteam.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from typing import Callable, Dict, List, Optional, Type

from pydantic.error_wrappers import ValidationError

from pybotx_smartapp_rpc.empty_args import EmptyArgs
from pybotx_smartapp_rpc.middlewares.empty_args_middleware import empty_args_middleware
from pybotx_smartapp_rpc.models.method import RPCMethod
from pybotx_smartapp_rpc.models.request import RPCRequest
from pybotx_smartapp_rpc.models.responses import (
    build_invalid_rpc_args_error_response,
    build_method_not_found_error_response,
)
from pybotx_smartapp_rpc.smartapp import SmartApp
from pybotx_smartapp_rpc.typing import (
    Handler,
    Middleware,
    RPCArgsBaseModel,
    RPCResponse,
)


class RPCRouter:
    def __init__(self, middlewares: Optional[List[Middleware]] = None) -> None:
        self.rpc_methods: Dict[str, RPCMethod] = {}
        self.middlewares: List[Middleware] = middlewares or []

    def method(
        self,
        rpc_method_name: str,
        middlewares: Optional[List[Middleware]] = None,
    ) -> Callable[[Handler], Handler]:
        if rpc_method_name in self.rpc_methods:
            raise ValueError(f"RPC method {rpc_method_name} already registered!")

        method_and_router_middlewares = self.middlewares + (middlewares or [])
        method_and_router_middlewares += [empty_args_middleware]

        def decorator(handler: Handler) -> Handler:
            annotations = list(handler.__annotations__.values())
            arguments_class: Optional[Type[RPCArgsBaseModel]] = None
            if len(annotations) == 3:
                # __annotations__ contains args and return typing
                # so len 3 means that method has rpc args
                arguments_class = annotations[1]

            self.rpc_methods[rpc_method_name] = RPCMethod(
                handler=handler,
                arguments_class=arguments_class,
                middlewares=method_and_router_middlewares,
            )

            return handler

        return decorator

    async def perform_rpc_request(
        self,
        smartapp: SmartApp,
        rpc_request: RPCRequest,
    ) -> RPCResponse:
        rpc_method = self.rpc_methods.get(rpc_request.method)
        if not rpc_method:
            return build_method_not_found_error_response(rpc_request.method)

        if rpc_method.arguments_class:
            try:
                args = rpc_method.arguments_class(**rpc_request.params)
            except ValidationError as invalid_rpc_args_exc:
                return build_invalid_rpc_args_error_response(invalid_rpc_args_exc)
        else:
            args = EmptyArgs()

        return await rpc_method(smartapp, args)

    def include(self, *routers: "RPCRouter") -> None:
        for router in routers:
            self.include_router(router)

    def include_router(self, router: "RPCRouter") -> None:
        already_exist_handlers = self.rpc_methods.keys() & router.rpc_methods.keys()
        if already_exist_handlers:
            raise ValueError(
                f"RPC methods {already_exist_handlers} already registered!",
            )

        for rpc_method_name, rpc_method in router.rpc_methods.items():
            rpc_method.middlewares = self.middlewares + rpc_method.middlewares
            self.rpc_methods[rpc_method_name] = rpc_method
